; -----------------------------------------------------------------------------
; RetroForth for "Generic" Systems
; -----------------------------------------------------------------------------
include "macros.fasm"

format elf


extrn putchar
extrn getchar
extrn exit
extrn open
extrn close
extrn read
extrn write
extrn chmod
extrn dlopen
extrn dlsym

public _start

_start:
	call init_retroforth
	call patch_vectors
	jmp boot
;---------------------------------------------------------------------
patch_vectors:
        upsh __use
        embed 'is use'
        upsh __read
        embed 'is r'
        upsh __write
        embed 'is w'
next
;---------------------------------------------------------------------
fname rb 256
erase_it:
        upsh fname
        embed '256 0 fill'
        next

__use:
        call erase_it
        upsh 32
        call parse
        upsh fname
        swap
        embed 'move'
        next

__nblk: embed '#blks @ 512 *'
        ret

__file dd 0
__open:
        dup
        push 66
        push fname
        call open
        add esp, 8
        mov [__file], eax
        push 438                ; 0666 in octal
        push fname
        call chmod
        drop
        add esp, 8
next
__close:
        push dword [__file]
        dup
        call close
        add esp, 4
        drop
next
__read:
        dup
        call __open
        call __nblk
        upop ebx
        push ebx
        push blk
	embed 'offset @ +'
        push dword [__file]
        call read
        add esp, 12
        call __close
        drop
next
__write:
        dup
        call __open
        call __nblk
        upop ebx
        push ebx
        push blk
        push dword [__file]
        call write
        add esp, 12
        call __close
        drop
next

code 'load', __load
	call __use
	jmp __read
code 'bye', bye
	vector
	push 0
	call exit

sys_emit:
	pusha
	push eax
	call putchar
	pop eax
	popa
	drop
ret

sys_key:
	dup
	call getchar
	ret

sys_type:
	push ebx	        ;
	push ecx	        ;
	upop ecx	        ;
	upop ebx	        ;
	or ecx,ecx	        ; Is the count zero?
	jz .done	        ; If so, we can exit
.a:	push ecx	        ; store the count
	dup		        ; Duplicate it
	xor eax,eax	        ; Filter out junk bits
	mov al,[ebx]	        ; Obtain the byte to show
	inc ebx 	        ; increment to the next byte
	call emit	        ; Show the character
	pop ecx 	        ; Restore the count
	loop .a 	        ; Loop, decrementing the count
.done:	pop ecx 	        ;
	pop ebx 	        ;
next
;---------------------------------------------------------------------
old_esp  dd 0
lib_addr dd 0

code 'from', _from
        upsh 32
        call parse
        embed 'zt'
        upop edx
        pusha
	push 1
	push edx
	call dlopen
	add esp, 8
        mov dword [lib_addr], eax
        popa
next

code 'import', _import
        pusha
        upsh 32
        call parse
        embed '2dup entry zt'
        upop edx
	push edx
	push [lib_addr]
	call dlsym
	add esp, 8
	cmp eax, 0
	jz import_error
        call literal
        popa
        call literal
        upsh _invoke
        call compile
        jmp ssemi

import_error:
	popa
	drop
	upsh error_message
	upsh 43
	call type
	embed 'last @ 9 + dup 1- c@ type cr last @ @ last !'
	ret
error_message db 'Error: Attempt to import unknown function: '


_invoke:
	mov dword [old_esp], esp
        swap
        upop [faddr]            ; Get the function address
        upop ecx
        cmp ecx, 0
        jz .run
.chain: push eax
        drop
        loop .chain
.run:   upsh [faddr]            ; Restore the function address
        call dword eax
	mov esp, dword [old_esp]
next
faddr dd 0
tempX dd 0
;---------------------------------------------------------------------
include 'core.asm'		; RetroForth itself
